/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#define _GNU_SOURCE
#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/prctl.h>
#include <netinet/in.h>
#include <errno.h>

#define PID_BUF_SIZE 32
#define PID_FILE "/var/run/zmd.pid"
#define REMOTING_SOCKET "/var/run/zmd/zmd-remoting.socket"
#define CONF_BUF_SIZE 256
#define CONF_FILE SYSCONFDIR "/zmd/zmd.conf"

static int
check_existing_zmd (void)
{    
    char buf[PID_BUF_SIZE];
    FILE *pid_stream;
    int pid;

    pid_stream = fopen (PID_FILE, "r");

    if (!pid_stream)
        return 0;

    if (!fgets (buf, PID_BUF_SIZE, pid_stream)) {
        fclose (pid_stream);
        return 0;
    }

    fclose (pid_stream);

    pid = atoi (buf);

    if (pid <= 0)
        return 0;

    return kill (pid, 0) == 0;
}

static char *
get_conf_value (const char *name)
{
    FILE *conf_stream;
    char buf[CONF_BUF_SIZE];
    int port = 2544;
    char *namestr;
    char *value = NULL;

    conf_stream = fopen(CONF_FILE, "r");

    if (!conf_stream)
        return NULL;

    asprintf (&namestr, "%s=", name);

    while (fgets (buf, CONF_BUF_SIZE, conf_stream)) {
        if (strstr (buf, namestr)) {
            value = strdup (buf + strlen (namestr));
            value[strlen(value) - 1] = '\0';
            break;
        }
    }

    free (namestr);
    fclose (conf_stream);

    return value;
}

static int
get_tcp_web_port (void)
{
    char *val;
    int port = 0;

    val = get_conf_value ("remote-port");
    if (val != NULL) {
        port = atoi (val);
    }

    if (port == 0) {
        return 2544;
    } else {
        return port;
    }
}

static char *
get_tcp_web_address (void)
{
    return get_conf_value ("bind-ip");
}

static int
create_tcp_web_socket (void)
{
    char *address;
    struct sockaddr_in sin;
    int fd;
    int opt;

    bzero(&sin, sizeof (sin));
    sin.sin_family = AF_INET;
    sin.sin_port = htons (get_tcp_web_port ());
    sin.sin_addr.s_addr = INADDR_ANY;

    address = get_tcp_web_address ();
    if (address != NULL && !inet_aton (address, &sin.sin_addr.s_addr)) {
        fprintf (stderr, "\nWARNING: could not bind to '%s'\n", address);
        sin.sin_addr.s_addr = INADDR_ANY;
    } else if (address == NULL) {
        sin.sin_addr.s_addr = INADDR_ANY;
    }

    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        fprintf (stderr, "\nWARNING: could not create tcp socket: %s\n", strerror (errno));
        return errno;
    }

    opt = 1;
    setsockopt (fd, SOL_SOCKET,
                SO_REUSEADDR, (void *) &opt, sizeof (opt));


    if (bind (fd, (struct sockaddr *) &sin, sizeof (sin)) < 0) {
        fprintf (stderr, "\nWARNING: unable to bind to tcp socket: %s\n", strerror (errno));
        return errno;
    }
    
    if (listen (fd, 5) < 0) {
        fprintf (stderr, "\nWARNING: unable to listen on tcp socket: %s\n", strerror (errno));
        return errno;
    }

    return 0;
}

static int
create_remoting_socket (void)
{
    char *dir;
    struct sockaddr_un sun;
    int fd;

    bzero (&sun, sizeof (sun));
    sun.sun_family = AF_UNIX;
    sprintf (sun.sun_path, REMOTING_SOCKET);

    if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
        fprintf (stderr, "\nERROR: could not create remoting socket: %s\n", strerror (errno));
        return errno;
    }

    unlink (REMOTING_SOCKET);
    
    if (bind (fd, (struct sockaddr *) &sun, sizeof (sun)) < 0) {
        fprintf (stderr, "\nERROR: unable to bind to remoting socket: %s\n", strerror (errno));
        return errno;
    }
    
    if (listen (fd, 5) < 0) {
        fprintf (stderr, "\nERROR: unable to listen on remoting socket: %s\n", strerror (errno));
        return errno;
    }

    chmod (REMOTING_SOCKET, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP |
           S_IWGRP | S_IXGRP | S_IROTH | S_IWOTH | S_IXOTH);
    
    return 0;
}

static int
create_sockets (int no_remote)
{
    int ret = 0;

    if ((ret = create_remoting_socket ()) != 0)
        return ret;

    if (!no_remote) {
        /* if it fails that's ok */
        create_tcp_web_socket ();
    }
    
    return 0;
}

int
main (int argc, char *argv[])
{
    int i = 0;
    int daemonize = 1;
    int min_args;
    int sleep = 0;
    int sleep_interval = -1;
    int first_time;
    int no_remote = 0;

#if HAVE_PR_SET_NAME
    prctl(PR_SET_NAME, "zmd", 0, 0, 0);
    argv[0] = "zmd";
#endif

    first_time = getenv ("ZMD_RUNNING") == NULL;
    
    min_args = 2; /* zmd-bin <assembly> <args> */

    if (argc < min_args) {
        fprintf (stderr, "\nERROR: Not enough arguments specified\n");
        exit (4);
    }
    
	
    if (getuid () != 0) {
        fprintf (stderr, "\nERROR: ZMD can only be run by root\n");
        exit (4);
    }

    if (first_time && check_existing_zmd ()) {
        /* fprintf (stderr, "An instance of ZMD is already running\n"); */
        exit (0);
    }

    for (i = 1; i < argc; i++) {

        if (strcmp (argv[i], "-n") == 0 ||
            strcmp (argv[i], "--no-daemon") == 0 ||
            strcmp (argv[i], "-h") == 0 ||
            strcmp (argv[i], "--help") == 0) {
            daemonize = 0;
        } else if (strcmp (argv[i], "--sleep") == 0) {
            argv[i] = "--sleep-resume";
            sleep = 1;

            if (argv[i+1] != NULL) {
                sleep_interval = atoi (argv[i+1]);
                argv[i+1] = "";
            }
        } else if (strcmp (argv[i], "-r") == 0 ||
                   strcmp (argv[i], "--no-remote") == 0) {
            no_remote = 1;
        }
    }

    if (daemonize && first_time) {
        int i = 0;
        int fd;

        /* this forks, exits the parent, and chdirs to / */
        daemon (0, 1);

        setsid ();
        umask (022);

        /* close all file descriptors */
        for (i = getdtablesize (); i >= 0; --i)
            close (i);

        /* reopen stdin to /dev/null */
        fd = open ("/dev/null", O_RDWR);
        assert (fd == STDIN_FILENO);

        /* redirect stdout to log file */
        fd = open (LOGDIR "/zmd-messages.log", O_APPEND | O_CREAT | O_WRONLY);
        assert (fd == STDOUT_FILENO);

        fd = dup (fd); /* dup stdout to stderr */
        assert (fd == STDERR_FILENO);
    }

    if (first_time) {
        char *current_time;
        char *remote_enabled;

        remote_enabled = get_conf_value ("remote-enabled");
        if (remote_enabled != NULL &&
            strcmp (remote_enabled, "False") == 0) {
            no_remote = 1;
        }
        
        int ret = create_sockets (no_remote);
        if (ret != 0)
            return ret;

        asprintf (&current_time, "%d", time(NULL));
        setenv ("ZMD_RUNNING", current_time, 1);
        free (current_time);
    }
    
    if (sleep) {
        struct timeval tv;
        fd_set rfds;
        int i;

        /* close any fds greater than the ones we're monitoring. */
        /* sadly, this does not take care of any 'stale' wapi handles */
        for (i = 5; i < getdtablesize (); i++) {
            close (i);
        }
        
	    /* monitor fds and wake up if needed */

        FD_ZERO (&rfds);
        FD_SET (3, &rfds); /* unix remoting socket */
        FD_SET (4, &rfds); /* tcp web server socket */
        
        if (sleep_interval > 0) {
            tv.tv_sec = sleep_interval;
            tv.tv_usec = 0;
            
            select (5, &rfds, NULL, NULL, &tv);
        } else {
            select (5, &rfds, NULL, NULL, NULL);
        }
    }

    unlink ("/var/run/zmd/zmd.sleeping");
    return execv (MONO, argv);
}
